// Copyright (c) 2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __01_GRAPH_HPP__
#define __01_GRAPH_HPP__

#include <cstring>

#include <rfdk/Graph.hpp>
#include <rfdk/utils.hpp>

namespace MRF{
	class ZeroOneGraph : public Graph{ // this constructs a 0-1 Graph, as described in: Wille, Anja and Bühlmann, Peter, "Low-Order Conditional Independence Graphs for Inferring Genetic Networks", 2005.
		private:
			using Graph::neighbors;
			using Graph::n;
			using Graph::m;
			using Graph::edges;
			using Graph::keepEdges;
			using Graph::tree;
			using Graph::colors;

			using Graph::nN;
			using Graph::nNR;

			double alpha;
			double minFreq;

			double  *m1;
			double **m2;
			size_t N;

			void init(){
				initGraph();
			}

			std::vector<std::string> fields;

		public:

			ZeroOneGraph(std::vector<std::string>& vN, std::string& filename, double a = 0.0, double mF = 0.0) : Graph() {

				minFreq = mF;
				alpha = a;

				n = vN.size();

				for(size_t i=0; i<n; ++i){
					nNR[vN[i]] = i+1;
					nN[i+1] = vN[i];
				}

				tree = false;

				N = 0;

				m1 = new double[n];
				m2 = new double*[n];
				for(size_t i=0; i<n; ++i) m2[i] = new double[n];

				for(size_t i=0; i<n; ++i)
					for(size_t j=0; j<n; ++j)
						m2[i][j] = 0.0;


				//std::cout << "LOADING DATA FOR 0-1 GRAPH" << std::endl;

				std::ifstream inRAW(filename);
				std::string line, header;

				getline(inRAW, header); // assume first line is header

				// read header
				size_t pos = header.find(','), lpos = 0;
				while(pos != std::string::npos){
					std::string name = header.substr(lpos, pos-lpos);
					fields.push_back(name);
					//std::cout << name << std::endl;
					lpos = pos+1;
					pos = header.find(',', lpos);
				}
				std::string name = header.substr(lpos);
				fields.push_back(name);
				//std::cout << name << std::endl;

				//std::cout << "FOUND HEADER FOR 0-1 GRAPH" << std::endl;

				// read data
				while(getline(inRAW, line)){

					// load current line into map
					std::map<std::string, double> current;
					size_t i = 0;
					pos = line.find(','), lpos = 0;
					while(pos != std::string::npos){
						std::string sval = line.substr(lpos, pos-lpos);
						assert(sval.size());

						std::stringstream vals;
						vals << sval;
						double val;
						vals >> val;
						current[fields[i]] = val;

						lpos = pos+1;
						pos = line.find(',', lpos);
						++i;
					}
					std::string sval = line.substr(lpos);
					assert(sval.size());

					std::stringstream vals;
					vals << sval;
					double val;
					vals >> val;
					current[fields[i]] = val;

					for(auto x1 : nNR){
						m1[x1.second-1] += current[x1.first];
						for(auto x2 : nNR){
							m2[x1.second-1][x2.second-1] += current[x1.first]*current[x2.first];
						}
					}

//					if(N%100 == 0) std::cout << ".(" << N << ")" << std::flush;
					if(N%100 == 0) std::cout << "." << std::flush;
					N++;
				}
				std::cout << std::endl;
				inRAW.close();

				//std::cout << "COMPUTED FIRST TWO MOMENTS FOR 0-1 GRAPH" << std::endl;

				for(auto x1 : nNR){
					m1[x1.second-1] /= N;
					for(auto x2 : nNR){
						m2[x1.second-1][x2.second-1] /= N;
					}
				}

				std::vector<std::string> vN2;

				for(unsigned int i=0; i<n; ++i){
					if(m1[i] > minFreq){
						vN2.push_back(vN[i]);
					}else{
						if(VERBOSE) std::cout << "SKIPPING VERTEX " << vN[i] << std::endl;
					}
				}

				n = vN2.size();
				nNR.clear();
				nN.clear();

				for(unsigned int i=0; i<n; ++i){
					nNR[vN2[i]]  = i+1;
					//std::cout << "nNR[" << vN2[i] << "] = " << nNR[vN2[i]] << std::endl;
					nN[i+1]  = vN2[i];
					//std::cout << "nN[" << i+1 << "] = " << nN[i+1] << std::endl;
				}

				for(unsigned int i=0; i<n; ++i){
					for(unsigned int j=i+1; j<n; ++j){

						bool nonzero = (std::abs(rho(i,j)) > alpha);

						for(unsigned int k=0; k<n; ++k){
							if(i!=k && j!=k && std::abs(rho(i,j,k)) <= alpha){
								nonzero = false;
								break;
							}
						}

						if(nonzero){
							edges->push_back(new _pair(i,j));
							//std::cout << "ADDING EDGE " << getVertexNameById(i) << "  " << getVertexNameById(j) << std::endl;
						}
					}
				}

				initGraph();
			}

			double mean(size_t i){
				return m1[i];
			}

			double var(size_t i, size_t j){
				return m2[i][j] - m1[i]*m1[j];
			}

			double rho(size_t i, size_t j){
				return var(i,j)/sqrt(var(i,i)*var(j,j));
			}

			double rho(size_t i, size_t j, size_t k){
				double rik = rho(i,k);
				double rjk = rho(j,k);

				return (rho(i,j)-rik*rjk)/sqrt((1-(rik*rik))*(1-(rjk*rjk)));
			}

			~ZeroOneGraph(){
				for(size_t i=0; i<n; ++i) delete[] m2[i];
				delete[] m2;
				delete[] m1;
			}
	};
};
#endif
