// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __ABSTRACT_MRF_HPP__
#define __ABSTRACT_MRF_HPP__

#include <iostream>
#include <iomanip>
#include <cmath>
#include <limits>
#include <cassert>
#include <cstring>

#include <rfdk/Graph.hpp>
#include <rfdk/Parameter.hpp>

#ifndef DEBUG
#define DEBUG false
#endif

namespace MRF{

#ifdef __NVCC__
	// vllt. extern templates..
	template <class R, bool doMAP> void computeMessagesLoopy_cuda(AbstractGraph*&, AbstractParameterStorage<R>*&, R*, R*, R*&, R*&, R*&, const unsigned int, const unsigned int, int*, int*&, R);
	template <class R> void free_cuda(R*&, R*&, R*&, int*&, AbstractGraph*&, AbstractParameterStorage<R>*&);
#endif

	template <class T> class AbstractMRF{
		protected:
#ifdef __NVCC__
			AbstractGraph* GG; // GPUgraph

			T* in_msgG;
			T* out_msgG;

			T* tRes;

			int* _OG;
#endif
			double* probs;

			AbstractParameterStorage<T>* par;

			Graph* G;

			T initMsg;

			T* in_msg;
			T* out_msg;
			T* prods;
			T* _Z;

			T alpha;

			unsigned int msgCount;
			unsigned int X;
			unsigned int dim;
			unsigned int mdim;
			unsigned int loopyI;

			unsigned int* _MAP; // map states
			int* _O; // observed states

			bool ownGraph;

			virtual void init(){
				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				msgCount = 0;

				dim  = n*X + m*X*X;
				mdim = 2*m*X;

				initMsg = 0;

				in_msg  = new T[mdim];
				out_msg = new T[mdim];
				_Z	= new T[n+m];
				prods	= new T[dim];

				probs   = new double[dim];
				_MAP	= new unsigned int[n];
				_O	= new int[n];

				clearObservations();

				memset(_Z, 0, (n+m)*sizeof(T));
				memset(prods, 0, dim*sizeof(T));
				memset(probs, 0, dim*sizeof(T));
			}

			inline __attribute__((always_inline)) unsigned int getClique(unsigned int i) const {
				const unsigned int n = G->getNumNodes();
				return ( (i<n*X) ? i/X : ((i-(n*X))/(X*X))+n );
			}

			inline __attribute__((always_inline)) unsigned int getState(unsigned int i) const {
				const unsigned int n = G->getNumNodes();
				return ( (i<n*X) ? i%X : (i-(n*X))%(X*X) );
			}

			void free(){
				if(X){
					delete[] in_msg;
					delete[] out_msg;
					delete[] probs;
					delete[] prods;
					delete[] _Z;
					delete[] _MAP;
					delete[] _O;
#ifdef __NVCC__
					free_cuda<T>(in_msgG, out_msgG, tRes, _OG, GG, par);
#endif
					//std::cout << "deleting AbstractMRF::par at " << par << std::endl;
					//delete par;
					if(ownGraph) delete G;
				}
			}

		private:

			virtual T accumulateProd(T a, T b) const  = 0;
			virtual T potential(T a, T b) const  = 0;
			virtual T result(T a) const  = 0;
			virtual double presult(T a, T b) const  = 0;

			virtual void preVertexProb(unsigned int){};
			virtual void postVertexProb(unsigned int){};
			virtual void preEdgeProb(unsigned int){};
			virtual void postEdgeProb(unsigned int){};

			virtual void postMessageProduct(unsigned int, unsigned int, T){};
			virtual void postMessageProduct(unsigned int, unsigned int, unsigned int l, T){};

			virtual void preMessage(unsigned int, bool){};
			virtual void postMessage(unsigned int, bool){};

			virtual T zresult(T a) const {
				return a;
			}

			virtual T edgezresult(T a) const {
				return a;
			}

			virtual T accumulateSum(T a, T b) const {
				return a + b;
			}

			virtual T accumulateZ(T a, T b) const {
				return a + b;
			}

			virtual T accumulateMax(T a, T b) const {
				return std::max(a, b);
			}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

			void err_lik() const {
				bool LIKELIHOOD_IS_NAN = true;
				assert(!LIKELIHOOD_IS_NAN);
			}

			void err_msg(unsigned int c, unsigned int x, T sold, T s) const {
				#pragma omp critical
				{
					std::cerr << "MESSAGE OVERFLOW AT EDGE " << c << std::endl;
					std::cerr << "STATE " << x << std::endl;
					std::cerr << "SUM " << s << std::endl;
					std::cerr << "OLD SUM " << sold << std::endl;
				}
				bool MESSAGE_OVERFLOW = true;
				assert(!MESSAGE_OVERFLOW);
			}

			void err_prob(unsigned int c, unsigned int x, double p) const {
				#pragma omp critical
				{
					std::cerr << "INVALID PROBABILITY AT COMPONENT " << c << std::endl;
					std::cerr << "STATE " << x << std::endl;
					std::cerr << "PROB " << p << std::endl;
				}
				bool INVALID_PROBABILITY = true;
				assert(!INVALID_PROBABILITY);
			}

			void err_z(unsigned int c, unsigned int x, T sold, T s) const {
				#pragma omp critical
				{
					std::cerr << "NORMALIZATION OVERFLOW AT COMPONENT " << c << std::endl;
					std::cerr << "STATE " << x << std::endl;
					std::cerr << "SUM " << s << std::endl;
					std::cerr << "OLD SUM " << sold << std::endl;
				}
				bool NORMALIZATION_OVERFLOW = true;
				assert(!NORMALIZATION_OVERFLOW);
			}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

			void computeProbs(){
				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();
#pragma omp parallel for
				for(unsigned int i=0; i<n; ++i){
					preVertexProb(i);
					for(unsigned int k = 0; k<X; ++k){
						prods[i*X+k] = prod(i, k);
					}
					postVertexProb(i);
					_Z[i] = prodZ(i);
					const double z = _Z[i];
					assert(z == z);
					assert(z != 0);
					for(unsigned int k = 0; k<X; ++k){
						T t = prods[i*X+k];
						assert(t == t);
						const double p = probs[i*X+k] = presult(t, z);
						if(p!=p || p<0 || p>1) err_prob(i,k,p);
						postMessageProduct(i, k, t);
					}
				}
#pragma omp parallel for
				for(int j=0; j<m; ++j){
					preEdgeProb(j);
					for(int k = 0; k<X; ++k){
						for(int l = 0; l<X; ++l){
							double pr = prodEdge(j, k, l);
							prods[n*X + j*X*X + k*X + l] = pr;
							assert(pr==pr);
							//assert(pr!=0);
						}
					}
					postEdgeProb(j);
					_Z[n+j] = prodEdgeZ(j);
					const double z = _Z[n+j];
					assert(z == z);
					assert(z != 0);
					for(int k = 0; k<X; ++k){
						for(int l = 0; l<X; ++l){
							T t = prods[n*X + j*X*X + k*X + l];
							assert(t==t);
							const double p = probs[n*X + j*X*X + k*X + l] = presult(t, z);
							if(p!=p || p<0 || p>1) err_prob(n+j,k*X+l,p);
						}
					}
				}
			}

			T prodZ(unsigned int s){
				T sum = 0, sum_old;

				for(unsigned int k=0; k<X; ++k){
					sum_old = sum;
//					T a = prod(s, k);
					T a = prods[s*X+k];
					sum = accumulateZ(sum, a);

					if(sum==0 || sum!=sum || sum < sum_old) err_z(s, k, sum_old, sum);
				}

				return zresult(sum);
			}

			T prod(unsigned int s, unsigned int x){
				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				T _prod = initMsg;
				for(auto ii = G->getNeighbors(s)->begin(); ii != G->getNeighbors(s)->end(); ++ii){
					T in = initMsg;
					if(G->getEdgeSource(ii->second) == s) in = in_msg[m*X + ii->second*X+x];
					else in = in_msg[ii->second*X+x];

					_prod = accumulateProd(_prod, in);
				}
				T val = potential(getLogPotential(s, x), _prod);
				//if(s==0) std::cerr << "PROD = " << val << std::endl;
				return val;
			}

			T prodEdgeZ(unsigned int eid){
				unsigned int s = G->getEdgeSource(eid);
				unsigned int t = G->getEdgeTarget(eid);

				const unsigned int n = G->getNumNodes();

				T sum = 0, sum_old;
				for(unsigned int k=0; k<X; ++k){
					for(unsigned int l=0; l<X; ++l){
						sum_old = sum;
						T a = prods[n*X + eid*X*X + k*X + l];
						sum = accumulateZ(sum, a);
						if(sum==0 || sum!=sum || sum < sum_old) err_z(n+eid, k*X+l, sum_old, sum);
					}
				}
				return edgezresult(sum);
			}

			T prodEdge(unsigned int eid, unsigned int x, unsigned int y){

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				unsigned int s = G->getEdgeSource(eid);
				unsigned int t = G->getEdgeTarget(eid);

				T _prod = initMsg;
				for(auto ii = G->getNeighbors(s)->begin(); ii != G->getNeighbors(s)->end(); ++ii){
					if(ii->first != t){
						T in = initMsg;
						if(G->getEdgeSource(ii->second) == s) in = in_msg[m*X + ii->second*X+x];
						else in = in_msg[ii->second*X+x];

						assert(in==in);
						//assert(in!=0);

						_prod = accumulateProd(_prod, in);
					}
				}
				for(auto ii = G->getNeighbors(t)->begin(); ii != G->getNeighbors(t)->end(); ++ii){
					if(ii->first != s){
						T in = initMsg;
						if(G->getEdgeSource(ii->second) == t) in = in_msg[m*X + ii->second*X+y];
						else in = in_msg[ii->second*X+y];

						assert(in==in);
						//assert(in!=0);

						_prod = accumulateProd(_prod, in);
					}
				}
				T res = potential(getLogPotential(eid, x, y) + getLogPotential(s, x) + getLogPotential(t, y), _prod);

				assert(res==res);
/*
				if(res==0){
					std::cout << "RESULT = " << res << std::endl;
					std::cout << "G->getNeighbors(s)->size() = " << G->getNeighbors(s)->size() << std::endl;
					std::cout << "G->getNeighbors(t)->size() = " << G->getNeighbors(t)->size() << std::endl;
					std::cout << "getLogPotential(" << eid << ", " << x << ", " << y << ") = " << getLogPotential(eid, x, y) << std::endl;
					std::cout << "getLogPotential(" << s << ", " << x << ") = " << getLogPotential(s, x) << std::endl;
					std::cout << "getLogPotential(" << t << ", " << y << ") = " << getLogPotential(t, y) << std::endl;
					std::cout << "_prod = " << _prod << std::endl;
					//assert(false);
				}
*/
				return res;
			}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

			void propagate(bool doMAP){
				initMessages();
				if(G->isTree()) computeMessages(doMAP);
#ifdef __NVCC__
				else{
					if(doMAP) computeMessagesLoopy_cuda<T, true>(GG, par, in_msg, out_msg, in_msgG, out_msgG, tRes, X, loopyI, _O, _OG, alpha);
					else computeMessagesLoopy_cuda<T, false>(GG, par, in_msg, out_msg, in_msgG, out_msgG, tRes, X, loopyI, _O, _OG, alpha);
				}
#else
				else computeMessagesLoopy(doMAP);
#endif
			}

			void propagate(){
				propagate(false);
			}

			void propagateMAP(){
				propagate(true);
			}

			void computeMAP(){

				const unsigned int n = G->getNumNodes();
#pragma omp parallel for
				for(unsigned int i=0; i<n; ++i){
					_MAP[i] = computeMAP(i);
				}
			}

			unsigned int computeMAP(unsigned int s){

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				assert(s>=0 && s<n);

				if(_O[s]!=-1) return _O[s];

				T r = 0;
				unsigned int rID = 0;
				for(unsigned int k=0; k<X; ++k){
					T logprod = initMsg;
					for(auto ii = G->getNeighbors(s)->begin(); ii != G->getNeighbors(s)->end(); ++ii){
						T in = initMsg;
						if(G->getEdgeSource(ii->second) == s) in = in_msg[m*X + ii->second*X+k];
						else in = in_msg[ii->second*X+k];
						logprod = accumulateProd(logprod, in);
					}

					T temp = potential(getLogPotential(s, k), logprod);
					if(temp > r){
						r = temp;
						rID = k;
					}
				}
				return rID;
			}

			void initMessages(){
				for(unsigned int i=0; i<mdim; ++i){
					in_msg[i]  = initMsg;
					out_msg[i] = initMsg;
				}
			}

			void computeMessagesLoopy(bool doMAP){

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				double eps = 0.01;
//				double eps = 0.000001; Paper

				double d = 1;

				for(unsigned int j=0; j<loopyI && d>eps; ++j){
#pragma omp parallel for
					for(unsigned int i=0; i<m; ++i){
						preMessage(i, false);
						for(unsigned int k=0; k<X; ++k){
							const unsigned int idx = i*X+k;
							out_msg[idx] = msg(i, k, doMAP, false);	// native edge direction
						}
						preMessage(i, true);
						for(unsigned int k=0; k<X; ++k){
							const unsigned int idx = m*X + i*X+k;
							out_msg[idx] = msg(i, k, doMAP, true);	// reverse edge direction
						}
					}
#pragma omp parallel for
					for(unsigned int i=0; i<m; ++i){
						postMessage(i, false);
						postMessage(i, true);
					}

					d = l2diff(in_msg, out_msg, 2*m*X);

					if(MRF::VERBOSEBP) std::cout << j << "\tdiff = " << d << std::endl;
				}
			}

			void computeMessages(bool doMAP){

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				for(unsigned int c=1; c<G->getNumColors()+1; ++c){
					for(unsigned int i=0; i<m; ++i){
						if(G->getColor(G->getEdgeSource(i))==c){
							preMessage(i, false);
							for(unsigned int k=0; k<X; ++k){
								out_msg[i*X+k] = msg(i, k, doMAP, false);	// native edge direction
							}
							postMessage(i, false);
						}
						else if(G->getColor(G->getEdgeTarget(i))==c){
							preMessage(i, true);
							for(unsigned int k=0; k<X; ++k){
								out_msg[m*X + i*X+k] = msg(i, k, doMAP, true);	// reverse edge direction
							}
							postMessage(i, true);
						}
					}
				}
				for(unsigned int c=G->getNumColors(); c>0; --c){
					for(unsigned int i=0; i<m; ++i){
						if(G->getColor(G->getEdgeSource(i))==c){
							preMessage(i, false);
							for(unsigned int k=0; k<X; ++k){
								//double tm = msg(i, k, doMAP, false);	// native edge direction
								//assert(tm == tm);
								out_msg[i*X+k] = msg(i, k, doMAP, false);	// native edge direction;
								
							}
							postMessage(i, false);
						}
						else if(G->getColor(G->getEdgeTarget(i))==c){
							preMessage(i, true);
							for(unsigned int k=0; k<X; ++k){
								out_msg[m*X + i*X+k] = msg(i, k, doMAP, true);	// reverse edge direction
							}
							postMessage(i, true);
						}
					}
				}
			}

			T msg(unsigned int eid, unsigned int x, bool doMAP, bool reverse){
				unsigned int s, t;

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				if(!reverse){
					s = G->getEdgeSource(eid);
					t = G->getEdgeTarget(eid);

				}else{
					s = G->getEdgeTarget(eid);
					t = G->getEdgeSource(eid);
				}
				if(DEBUG) std::cerr << "COMPUTING MESSAGE {" << s << " -> " << t << "}(" << x << ")" << std::endl;

				const int o = _O[s]; // load observed value (is -1 if node is unobserved)

				T _prod, sum = 0, sum_old;
				if(o == -1){
					for(unsigned int k=0; k<X; ++k){
						msgCount++;
						T theta_s, theta_st;
						_prod = initMsg;
					
						for(auto ii = G->getNeighbors(s)->begin(); ii != G->getNeighbors(s)->end(); ++ii){
							if(ii->first != t){
								T in;
								if(G->getEdgeSource(ii->second) == s) in = in_msg[m*X + ii->second*X+k];
								else in = in_msg[ii->second*X+k];

								if(DEBUG) std::cerr << "INBOUND {" << ii->first << " -> " << s << "}(" << k << ") = " << in << std::endl;
								_prod = accumulateProd(_prod, in);
							}
						}
						theta_s = getLogPotential(s, k);
					
						if(!reverse)	theta_st = getLogPotential(eid, k, x);
						else		theta_st = getLogPotential(eid, x, k);

						if(DEBUG){
							std::cerr << "WEIGHT(" << s << "=" << k << ") = " << (theta_s) << std::endl;
							std::cerr << "WEIGHT(" << s << "=" << k << ", " << t << "=" << x << ") = " << (theta_st) << std::endl;
						}
						sum_old = sum;

						if(doMAP)	sum = accumulateMax(sum, potential(theta_s+theta_st, _prod)); // Max-Product
						else		sum = accumulateSum(sum, potential(theta_s+theta_st, _prod)); // Sum-Product

						if(sum < sum_old) err_msg(eid, x, sum_old, sum);
					}
				}else{
					T theta_st;

					if(!reverse)	theta_st = getLogPotential(eid, o, x);
					else		theta_st = getLogPotential(eid, x, o);

					sum = potential(getLogPotential(s, o)+theta_st, initMsg);
				}
				if(DEBUG) std::cerr << "RESULT {" << s << " -> " << t << "}(" << x << ") = " << result(sum) << std::endl << std::endl;
				return result(sum);
			}

		public:
			AbstractMRF(){
				// can only be initialized from file or derived class

				X	= 0;
				G	= 0;
				in_msg	= 0;
				out_msg = 0;
				par	= 0;
				probs	= 0;
				_Z	= 0;
				_MAP	= 0;
				_O	= 0;
				ownGraph= 0;
				alpha	= 0;
				loopyI = -1;
			}

			AbstractMRF(Graph* _G, unsigned int _numStates, bool createParameterStorage = true){
				X = _numStates;
				G = _G;
#ifdef __NVCC__
				GG = _G;
				par = 0;
				in_msgG = 0;
				out_msgG = 0;
				tRes = 0;
				_OG = 0;
				alpha = 0;
#endif
				ownGraph = false;
				loopyI = -1;
				init();
				if(createParameterStorage){
					par = new ParameterStorage<T>(dim);
				}
			}

			~AbstractMRF(){
				free();
			}

			void setParameter(AbstractParameterStorage<T>* _P){
				par = _P;
				//std::cout << "SETTING AbstractMRF::par to " << par << std::endl;
			}

			virtual T optimize(T* F, unsigned int N) = 0;

			unsigned int getMessageCount(){
				return msgCount;
			}

			void resetMessageCount(){
				msgCount = 0;
			}

			virtual double logLikelihood(T* F, const unsigned int N) const {

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				double a = 0.0, b = 0.0, H = 0.0;

				for(unsigned int i=0; i<dim; ++i) a += (F[i]/N)*getWeight(i);
				if(a!=a) err_lik();

				for(unsigned int i=0; i<dim; ++i) b += probs[i]*getWeight(i);
				if(b!=b) err_lik();

				for(unsigned int i=0; i<n; ++i){
					long val = G->getNeighbors(i)->size();
					for(unsigned int j=0; j<X; ++j){
						const unsigned int idx = i*X+j;
						double t = probs[idx] * std::log(probs[idx]);
						H -= t;
					}
				}
				if(H!=H) err_lik();

				for(unsigned int i=0; i<m; ++i){
					for(unsigned int j=0; j<X; ++j){
						for(unsigned int k=0; k<X; ++k){
							const unsigned int idx = n*X + i*X*X + j*X + k;
							const unsigned int v = G->getEdgeSource(i);
							const unsigned int u = G->getEdgeTarget(i);
							double t = probs[idx] * (std::log(probs[idx]) - std::log(probs[v*X+j]) - std::log(probs[u*X+k]) );
							H += t;
						}
					}
				}
				if(H!=H) err_lik();

				//std::cout << "A(w) = " << (b+H) << std::endl;
				return a-(b+H);
			}

			double log2Likelihood(T* F, const unsigned int N) const {

//				double eps = 0.00001;

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				double result = 0.0;

				for(unsigned int i=0; i<n; ++i){
					double val = G->getNeighbors(i)->size();
					double Z = 0.0, Q = 0.0;
					for(unsigned int j=0; j<X; ++j){
						double t = (double(F[i*X+j])/double(N))*(val-1.0)*(std::log(probs[i*X+j])/std::log(2.0));

						assert(probs[i*X+j]>0.0 && probs[i*X+j]<1.0);

						result -= t;
					}
				}

				for(unsigned int i=0; i<m; ++i){
					double Z = 0.0, Q = 0.0;
					for(unsigned int j=0; j<X; ++j){
						for(unsigned int k=0; k<X; ++k){
							const unsigned int idx = n*X + i*X*X + j*X + k;
							double t = (double(F[idx])/double(N))*(std::log(probs[idx])/std::log(2.0));

							assert(probs[idx]>0.0 && probs[idx]<1.0);

							result += t;
						}
					}
				}
				return result;
			}


			double entropy(T* F, const unsigned int N) const {

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				double result = 0.0;

				for(unsigned int i=0; i<n; ++i){
					double val = G->getNeighbors(i)->size();
					for(unsigned int j=0; j<X; ++j){
						const unsigned int idx = i*X+j;
						//std::cout << "F[" << idx << "] = " << F[idx] << std::endl;
						double t = probs[idx] * (val-1.0) * std::log(probs[idx]);
						assert(probs[idx]>0.0 && probs[idx]<1.0);
						result -= t;
					}
				}

				for(unsigned int i=0; i<m; ++i){
					for(unsigned int j=0; j<X; ++j){
						for(unsigned int k=0; k<X; ++k){
							const unsigned int idx = n*X + i*X*X + j*X + k;
							//std::cout << "F[" << idx << "] = " << F[idx] << std::endl;
							double t = probs[idx] * std::log(probs[idx]);
							assert(probs[idx]>0.0 && probs[idx]<1.0);
							result += t;
						}
					}
				}
				return -result;
			}

			double entropy2(T* F, const unsigned int N) const {

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				double result = 0.0;

				for(unsigned int i=0; i<n; ++i){
					double val = G->getNeighbors(i)->size();
					for(unsigned int j=0; j<X; ++j){
						const unsigned int idx = i*X+j;
						double t = probs[idx]*(val-1.0)*(std::log(probs[idx])/std::log(2.0));
						assert(probs[idx]>0.0 && probs[idx]<1.0);

						result -= t;
					}
				}

				for(unsigned int i=0; i<m; ++i){
					for(unsigned int j=0; j<X; ++j){
						for(unsigned int k=0; k<X; ++k){
							const unsigned int idx = n*X + i*X*X + j*X + k;
							double t = probs[idx]*(std::log(probs[idx])/std::log(2.0));
							assert(probs[idx]>0.0 && probs[idx]<1.0);

							result += t;
						}
					}
				}

				return -result;
			}

			double grad1norm(T* F, unsigned int N) const {
				double r = 0.0;
				for(unsigned int i=0; i<dim; ++i){
					double val = (F[i]/double(N))-probs[i];
					r += std::abs(val);
				}
				return r;
			}

			double grad2norm(T* F, unsigned int N) const {
				double r = 0.0;
				for(unsigned int i=0; i<dim; ++i){
					double val = (F[i]/double(N))-probs[i];
					r += val*val;
				}
				return std::sqrt(r);
			}


			virtual void marginals(){
				propagate();
				computeProbs();
			}

			virtual void MAP(){
				propagateMAP();
				computeMAP();
			}

			unsigned int* getMAP() const {
				return _MAP;
			}

			unsigned int getMAPState(unsigned int i) const {
				assert(i>=0 && i<G->getNumNodes());
				return _MAP[i];
			}

			double* getProbs() const {
				return probs;
			}

			T* getProds() const {
				return prods;
			}

			T* getWeights() const {
				return par->getArray();
			}

			Graph* getGraph() const {
				return G;
			}

			unsigned int getDim() const {
				return dim;
			}

			void setLoopyIterations(int i){
				loopyI = i;
			}

			void showVector(T* F){
				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				for(unsigned int i=0; i<dim; ++i){
					if(F[i]){
						std::cerr << i << '\t' << F[i] << "\t";
						if(i < n*X){
							unsigned int vid = i/X;
							unsigned int x = i-(vid*X);

							std::cerr << vid << " = " << x << std::endl;
						}else{
							unsigned int eid = (i-n*X)/(X*X);
							unsigned int t = ((i-n*X)-(eid*X*X));
							unsigned int x = t/X;
							unsigned int y = t%X;

							std::cerr << "(" << G->getEdgeSource(eid) << "," << G->getEdgeTarget(eid) << ") = (" << x << "," << y << ")" << std::endl;
						}
					}
				}
			}

			virtual void zeroWeights(){
				par->zero();
			}

			unsigned int getNNZ() const {
				unsigned int result = 0;
				for(unsigned int i=0; i<dim; ++i){
					if(getWeight(i) != 0){
						result++;
					}
				}
				return result;
			}

			T Z(unsigned int c) const {
				//assert(c>=0 && c<G->getNumNodes()+G->getNumEdges());
				return _Z[c];
			}

			void setObservation(unsigned int v, int x){
				assert(v>=0 && v<G->getNumNodes());
				_O[v] = x;
			}

			int getObservation(unsigned int v){
				return _O[v];
			}

			void clearObservations(){
				for(unsigned int i=0; i<G->getNumNodes(); ++i){
					_O[i] = -1;
				}
			}

			T getWeight(unsigned int* x){
				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				// generalize this for larger clique sizes!
				T result = 0;
				for(unsigned int v=0; v<n; ++v)
					result += getWeight(v, x[v]);

				for(unsigned int e=0; e<m; ++e)
					result += getWeight(e, x[G->getEdgeSource(e)], x[G->getEdgeTarget(e)]);

				return potential(result, 0);
			}

			inline __attribute__((always_inline)) T getWeight(const unsigned int i) const {
				assert(i>=0 && i<dim);
				return par->get(i);
			}

			inline __attribute__((always_inline)) T getWeight(const unsigned int s, unsigned int x) const {
				assert(s>=0 && s<G->getNumNodes() && x>=0 && x<X);
				return getWeight(s*X+x);
			}

			inline __attribute__((always_inline)) T getWeight(const unsigned int eid, const unsigned int x, const unsigned int y) const {
				assert(eid>=0 && eid<G->getNumEdges() && x>=0 && x<X && y>=0 && y<X);
				return getWeight(G->getNumNodes()*X+eid*X*X+x*X+y);
			}

			virtual T getLogPotential(const unsigned int i) const {
				return this->getWeight(i);
			}

			virtual T getLogPotential(const unsigned int s, const unsigned int x) const {
				return getLogPotential(s*X+x);
			}

			virtual T getLogPotential(const unsigned int eid, const unsigned int x, const unsigned int y) const {
				return getLogPotential(G->getNumNodes()*X+eid*X*X+x*X+y);
			}

			inline __attribute__((always_inline)) void setWeight(unsigned int i, T val){
				assert(i>=0 && i<dim);
				par->set(i, val);
			}

			inline __attribute__((always_inline)) void setWeight(unsigned int s, unsigned int x, T val){
				//assert(s>=0 && s<G->getNumNodes() && x>=0 && x<X);
				setWeight(s*X+x, val);
			}

			inline __attribute__((always_inline)) void setWeight(unsigned int eid, unsigned int x, unsigned int y, T val){
				//assert(eid>=0 && eid<G->getNumEdges() && x>=0 && x<X && y>=0 && y<X);
				setWeight(G->getNumNodes()*X+eid*X*X+x*X+y, val);
			}

			inline __attribute__((always_inline)) double getProb(const unsigned int i) const {
				//assert(i>=0 && i<dim);
				return probs[i];
			}

			inline __attribute__((always_inline)) double getProb(const unsigned int s, const unsigned int x) const {
				//assert(s>=0 && s<G->getNumNodes() && x>=0 && x<X);
				return probs[s*X+x];
			}

			inline __attribute__((always_inline)) double getProb(const unsigned int eid, const unsigned int x, const unsigned int y) const {
				//assert(eid>=0 && eid<G->getNumEdges() && x>=0 && x<X && y>=0 && y<X);
				return probs[G->getNumNodes()*X+eid*X*X+x*X+y];
			}


                        void setDamping(T a){
                                alpha = a;
                        }

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

			virtual void storeConfig(std::ostream& out) const {
				out << X << std::endl;
			}

			virtual void loadConfig(std::istream& in){
				in >> X;
			}

			void storeParameters(std::ostream& out) const {

				const unsigned int n = G->getNumNodes();

				unsigned int _dim = 0;

				for(unsigned int i=0; i<dim; ++i) if(getWeight(i) != 0) _dim++;

				out << _dim << std::fixed << std::setprecision(16) << std::endl;

				for(unsigned int i=0; i<dim; ++i){
					if(getWeight(i) != 0){
						out << i << '\t' << getWeight(i) << "# ";
						if(i < n*X){
							unsigned int vid = i/X;
							unsigned int x = i-(vid*X);

							out << vid << " = " << x << std::endl;
						}else{
							unsigned int eid = (i-n*X)/(X*X);
							unsigned int t = ((i-n*X)-(eid*X*X));
							unsigned int x = t/X;
							unsigned int y = t%X;

							out << "(" << G->getEdgeSource(eid) << "," << G->getEdgeTarget(eid) << ") = (" << x << "," << y << ")" << std::endl;
						}
					}
				}
			}

			void loadParameters(std::istream& in){

//				std::cout << "loadParameters(" << &in << ")" << std::endl;

				std::string s;
				getline(in, s);
				std::stringstream ssm(s);

				unsigned int _dim;
				ssm >> _dim;

				for(unsigned int i=0; i<_dim; ++i){

					getline(in, s);
					if(s.size()>2){
						const unsigned int tpos = s.find('\t');
						const unsigned int rpos = s.find('#');

						std::string a = s.substr(0, tpos);
						std::string b = s.substr(tpos+1, rpos);

						std::stringstream as(a);
						std::stringstream bs(b);

						unsigned int va;
						T vb;

						as >> va;
						bs >> vb;

						setWeight(va, vb);
					}
				}
			}

			friend std::istream& operator>> (std::istream& in, MRF::AbstractMRF<T>& mrf){
				mrf.loadConfig(in);
				mrf.G = new Graph(in);
				mrf.ownGraph = true;
				mrf.init();
				mrf.loadParameters(in);
				return in;
			}

			friend std::ostream& operator<< (std::ostream& out, const MRF::AbstractMRF<T>& mrf){
				mrf.storeConfig(out);
				out << (*mrf.G);
				mrf.storeParameters(out);
				return out;
			}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

	};
};
#endif
