// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __GPUGRAPH_HPP__
#define __GPUGRAPH_HPP__

#include <iostream>
#include <rfdk/AbstractGraph.hpp>

namespace MRF{

	class GPUGraph : public AbstractGraph{

		private:

			using AbstractGraph::n;
			using AbstractGraph::m;
			using AbstractGraph::tree;

			unsigned int*	numNeighbors;
			unsigned int*	edges;
			unsigned int**	neighbors;
			unsigned int**	neighborsH;

			unsigned int*	neighborhoodSizes;
			unsigned int*	neighborhoodSizeCounts;
			unsigned int**	neighborhoodSizeSets;

			unsigned int	numNeighborhoodSizes;

		public:

			GPUGraph(AbstractGraph* G){

				n = G->getNumNodes();
				m = G->getNumEdges();

////////////////////////////////////////////////////////////////////////////

				unsigned int* nNtemp = new unsigned int[n];
				cudaMalloc((void**)&numNeighbors, n*sizeof(unsigned int));

				for(unsigned int i=0; i<n; ++i){
					nNtemp[i] = 0;
				}

				unsigned int* eTemp = new unsigned int[2*m];
				cudaMalloc((void**)&edges, 2*m*sizeof(unsigned int));

				for(unsigned int i=0; i<m; ++i){
					unsigned int a = G->getEdgeSource(i);
					unsigned int b = G->getEdgeTarget(i);

					eTemp[i*2+0] = a;
					eTemp[i*2+1] = b;

					++nNtemp[a];
					++nNtemp[b];
				}

				cudaMemcpy(edges, eTemp, 2*m*sizeof(unsigned int), cudaMemcpyHostToDevice);
				delete[] eTemp;

				cudaMemcpy(numNeighbors, nNtemp, n*sizeof(unsigned int), cudaMemcpyHostToDevice);
				delete[] nNtemp;

////////////////////////////////////////////////////////////////////////////

				neighborsH = new unsigned int*[n];
				cudaMalloc((void**)&neighbors, n*sizeof(unsigned int*));

				for(unsigned int j=0; j<n; ++j){
					unsigned int num = G->getNumNeighbors(j);
					unsigned int N[2*num];
					for(unsigned int i=0; i<num; ++i){
						N[2*i+0] = G->getNeighbor(j,i);
						N[2*i+1] = G->getNeighborEdge(j,i);
					}
					cudaMalloc((void**)(&(neighborsH[j])), 2*num*sizeof(unsigned int*));
					cudaMemcpy(neighborsH[j], N, 2*num*sizeof(unsigned int), cudaMemcpyHostToDevice);
				}
				cudaMemcpy(neighbors, neighborsH, n*sizeof(unsigned int*), cudaMemcpyHostToDevice);

////////////////////////////////////////////////////////////////////////////

				numNeighborhoodSizes = G->getNumNeighborCounts();

				cudaHostAlloc((void**)&neighborhoodSizes, numNeighborhoodSizes*sizeof(unsigned int), cudaHostAllocMapped);
				cudaHostAlloc((void**)&neighborhoodSizeCounts, numNeighborhoodSizes*sizeof(unsigned int), cudaHostAllocMapped);
				cudaHostAlloc((void**)&neighborhoodSizeSets, numNeighborhoodSizes*sizeof(unsigned int*), cudaHostAllocMapped);

				for(unsigned int i=0; i<numNeighborhoodSizes; ++i){
					unsigned int N = neighborhoodSizes[i] = G->getNumNeighborSize(i); // what is neighborhood-size no. i?
					unsigned int K = neighborhoodSizeCounts[i] = G->getNumNeighborCounts(i); // how many nodes have neighborhood-size no. i?

					unsigned int cudeResultCode = cudaHostAlloc((void**)&neighborhoodSizeSets[i], N*K*sizeof(unsigned int*), cudaHostAllocMapped);
//					std::cout << "ALLOCATED " << N*K << " ENTRIES FOR VERTEX " << i << cudeResultCode << std::endl;
//					std::cout << "cudaHostAlloc((void**)&neighborhoodSizeSets[i], N*K*sizeof(unsigned int*), cudaHostAllocMapped) = " << cudeResultCode << std::endl;
					for(unsigned int j=0; j<K; ++j){ // for all nodes with N neighbors
						unsigned int s = G->getNodeWithNeighborCount(i,j);
						for(unsigned int k=0; k<N; ++k){ // for all neighbors of node j
							// what is the edge id?
							// what is the direction? (reverse or not?)
							unsigned int eid = G->getNeighborEdge(s,k);

							unsigned int a = G->getEdgeSource(eid);
//							unsigned int b = G->getEdgeTarget(eid);

							bool reverse = false;

							if(a!=s){
								reverse = true;
//								unsigned int temp = a;
//								a = b;
//								b = temp;
							}
//							std::cout << "WRITING ENTRY " << i << cudeResultCode << std::endl;
							neighborhoodSizeSets[i][j*N+k] = reverse*m + eid;
						}
					}
				}
			}

			// return the i-th "node" (eid) that has degree j
			virtual __host__ __device__ unsigned int getNodeWithNeighborCount(unsigned int j, unsigned int i) const{
				return neighborhoodSizeSets[j][i];
			}

			virtual __host__ __device__ unsigned int* getNodeWithNeighborCountArray(unsigned int j) const{
				return neighborhoodSizeSets[j];
			}

			// What is the degree no. j
			virtual __host__ __device__ unsigned int getNumNeighborSize(unsigned int j) const {
				return neighborhoodSizes[j];
			}
/*
			virtual __host__ __device__ unsigned int getMaxNeighborhoodSize() const {
				unsigned int m = 0;
				for(unsigned int i=0; i<getNumNeighborCounts(); ++i) m = max(m,neighborhoodSizes[i]);
				return m;
			}
*/
			void finish(){
				cudaFree(numNeighbors);
				cudaFree(edges);
				for(unsigned int i=0; i<n; ++i){
					cudaFree(neighborsH[i]);
				}
				cudaFree(neighbors);

				cudaFreeHost(neighborhoodSizes);
				cudaFreeHost(neighborhoodSizeCounts);
				for(unsigned int i=0; i<numNeighborhoodSizes; ++i){
					cudaFreeHost(neighborhoodSizeSets[i]);
				}
				cudaFreeHost(neighborhoodSizeSets);

				delete[] neighborsH;
			}

			virtual __host__ __device__ unsigned int getNumNeighborCounts() const {
				return numNeighborhoodSizes;
			}

			virtual __host__ __device__ unsigned int getNumNeighborCounts(unsigned int j) const {
				return neighborhoodSizeCounts[j];
			}

			virtual __host__ __device__ unsigned int getNumNeighbors(unsigned int j) const {
				return numNeighbors[j];
			}

			virtual __host__ __device__ unsigned int getEdgeSource(unsigned int j) const {
				return edges[j*2+0];
			}

			virtual __host__ __device__ unsigned int getEdgeTarget(unsigned int j) const {
				return edges[j*2+1];
			}

			virtual __host__ __device__ unsigned int getNeighbor(unsigned int j, unsigned int i) const {
				return neighbors[j][i*2+0];
			}

			virtual __host__ __device__ unsigned int getNeighborEdge(unsigned int j, unsigned int i) const {
				return neighbors[j][i*2+1];
			}
	};
};
#endif
