// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __GRAPH_HPP__
#define __GRAPH_HPP__

#include <set>
#include <map>
#include <sstream>
#include <cassert>

#include <rfdk/AbstractGraph.hpp>
#include <rfdk/utils.hpp>

namespace MRF{
	class Graph : public AbstractGraph{
		protected:
			std::vector<std::set<_pair>* >* neighbors;
			_vec* edges;

			using AbstractGraph::n;
			using AbstractGraph::m;
			using AbstractGraph::tree;

			unsigned int num_colors;

			unsigned int* colors;

			bool keepEdges;

			std::map<unsigned int, std::set<unsigned int> > neighborCounts;

			std::map<unsigned int,std::string> nN;  // mapping {0,1,..,n} -> {NAME0,NAME1,..,NAMEn}
			std::map<std::string,unsigned int> nNR; // mapping {NAME0,NAME1,..,NAMEn} -> {0,1,..,n}

			void loadGraph(std::istream& in){

				std::string s;

				//std::stringstream ms(s);
				n=0;

				while(getline(in, s)){

					size_t pos = s.find(',');
					if(pos == std::string::npos) break;
					n++;

					std::string _a = s.substr(0,pos);
					unsigned int a = str2uint(_a)+1;
					std::string b = s.substr(pos+1);

					nN[a]  = b;
					nNR[b] = a;
					//std::cout << a << '\t' << b << std::endl;
				}

				std::stringstream sstr(s);

				unsigned int _m;
				sstr >> _m;

				//std::cout << "#EDGES = " << _m << std::endl;

				for(unsigned int i=0; i<_m; ++i){
					getline(in, s);
//					std::cout << "LINE = " << s << std::endl;

					const unsigned int cpos = s.find(',');

					std::string a = s.substr(0, cpos);
					std::string b = s.substr(cpos+1, s.size()-1);

					std::stringstream as(a);
					std::stringstream bs(b);

					unsigned int va, vb;
					as >> va;
					bs >> vb;

					edges->push_back(new MRF::_pair(va, vb));
//					std::cout << i << " constructed edge at " << pp << std::endl;
				}
			}

			void storeGraph(std::ostream& out) const {

				for(unsigned int i=0; i<n; ++i){
					const std::string name = getVertexNameById(i);
					out << i << ',' << name << std::endl;
				}

				out << m << std::endl;

				for(unsigned int i=0; i<m; ++i){
					out << getEdgeSource(i) << ',' << getEdgeTarget(i) << std::endl;
				}
			}

			virtual void initGraph(){
				m = edges->size();

				neighbors = new std::vector<std::set<_pair>*>();
				for(unsigned int i=0; i<n; ++i){
					neighbors->push_back(new std::set<_pair>());
				}

				for(unsigned int i=0; i<m; ++i){
					neighbors->at(edges->at(i)->first)->insert(_pair(edges->at(i)->second, i));
					neighbors->at(edges->at(i)->second)->insert(_pair(edges->at(i)->first, i));
				}

				for(unsigned int i=0; i<n; ++i){
					neighborCounts[neighbors->at(i)->size()].insert(i);
				}

				//for(auto ii : neighborCounts){
				//	std::cout << ii.first << '\t' << (ii.second).size() << std::endl;
				//}
			}

			unsigned int colorLeafs(std::vector<std::set<_pair>* >* ne, unsigned int *colors, unsigned int c){
				unsigned int i = 0;
				unsigned int removed = 0;
				// change to c++11 foreach loop..
				for(auto ii=ne->begin(); ii!=ne->end(); ++i){
					//assert((*ii)->size() == ne->at(i)->size());
					if(VERBOSEGRAPH) std::cout << "SIZE("<<i<<") = "<< (*ii)->size() << std::endl;
					if((*ii)->size() == 1){
						colors[i] = c;
						if(VERBOSEGRAPH) std::cout << "assigned color " << c << " to vertex " << i << std::endl;
					}
					++ii;
				}
				if(VERBOSEGRAPH) std::cout << std::endl;
				i = 0;
				std::set<unsigned int> visited;
				for(auto ii=ne->begin(); ii!=ne->end(); ++i){
					//assert((*ii)->size() == ne->at(i)->size());

					if((*ii)->size() == 1 && visited.find(i) == visited.end()){

						for(auto jj=ne->at(i)->begin(); jj!=ne->at(i)->end(); ++jj){
							ne->at(jj->first)->erase(_pair(i,jj->second));
							visited.insert(jj->first);
							++removed;
						}

						(*ii)->clear();
						++removed;
					}
					++ii;
				}
				return removed;
			}

			virtual void init(){
				//n = 0;
				m = 0;
				num_colors = 0;

				initGraph();
				colors = new unsigned int[n];
				//computeColoring();
			}

		public:

			std::string getVertexNameById(const unsigned int id) const {
				auto x = nN.find(id+1);
				if(x != nN.end()) return nN.at(id+1);
				return "";
			}

			unsigned int getVertexIdByName(std::string& s){
				return nNR[s]-1;
			}

			virtual void computeColoring(){ // this is multiple-root top-sort
				for(unsigned int i=0; i<n; ++i) colors[i] = 0;

				std::vector<std::set<_pair>* >* ne = new std::vector<std::set<_pair>* >();
				for(auto N : *neighbors) ne->push_back(new std::set<_pair>(*N));

				unsigned int total = 0;
				for(auto N : *ne) total += N->size(); // sum of all neighborhood sizes

				num_colors = 0;
				while(total) total -= colorLeafs(ne, colors, ++num_colors);
				bool lastColor = false;
				for(unsigned int i=0; i<n; ++i) if(!colors[i]){
					lastColor = true;
					colors[i] = num_colors+1;
					if(VERBOSEGRAPH) std::cout << "assigned color " << (num_colors+1) << " to vertex " << i << std::endl;
				}
				if(VERBOSEGRAPH) std::cout << std::endl;
				if(lastColor) ++num_colors;

				for(auto ii = ne->begin(); ii != ne->end(); ++ii) delete *ii;
				delete ne;
			}

			void combine(Graph* _G){
				_vec* other = _G->getEdges();

				// copy and rename edges
				for(auto ii = other->begin(); ii != other->end(); ++ii){
					edges->push_back(new _pair((*ii)->first+n, (*ii)->second+n));
				}
				// connect random nodes from both graphs
				edges->push_back(new _pair(rand()%n, n+(rand()%_G->getNumNodes())));
				this->free();
				n += _G->getNumNodes();
			}

			void finalize(){
				init();
			}

			Graph(_vec* _edges){
				edges = _edges;
				tree = true;
				keepEdges = true;
				init();
			}

			Graph(std::istream& iG){
				edges = new MRF::_vec();
				tree = true;
				keepEdges = false;
				loadGraph(iG);
				init();
			}

			Graph(){
				edges = new MRF::_vec();
				tree = true;
				keepEdges = false;
				neighbors = 0;
				colors = 0;
			}

			void free(){

				if(neighbors){
					for(auto ii = neighbors->begin(); ii != neighbors->end(); ++ii){
						delete *ii;
					}
					delete neighbors;
					neighbors = 0;
				}

				if(colors){
					delete[] colors;
					colors = 0;
				}
			}

			~Graph(){
				this->free();

				if(!keepEdges){
					for(int i = 0; i<edges->size(); ++i){
//						std::cout << i << " deleting edge at " << edges->at(i) << std::endl;
						delete edges->at(i);
					}
					edges->clear();
					delete edges;
				}
			}

			virtual unsigned int getEdgeSource(unsigned int j) const {
				assert(j<m);
				return edges->at(j)->first;
			}

			virtual unsigned int getEdgeTarget(unsigned int j) const {
				assert(j<m);
				return edges->at(j)->second;
			}

			virtual unsigned int getNumColors() const {
				return num_colors;
			}

			virtual unsigned int getColor(unsigned int i) const {
				return colors[i];
			}

			unsigned int getNumNodes() const {
				return n;
			}

			virtual unsigned int getNumEdges() const {
				return m;
			}

			_vec* getEdges() const {
				return edges;
			}

			std::set<_pair>* getNeighbors(unsigned int i) const {
				return neighbors->at(i);
			}

			virtual unsigned int getNeighbor(unsigned int j, unsigned int i) const{
				auto ii = neighbors->at(j)->begin();
				int k = 0;
				while(k++ < i) ii++;
				return ii->first;
			}

			virtual unsigned int getNeighborEdge(unsigned int j, unsigned int i) const{
				auto ii = neighbors->at(j)->begin();
				int k = 0;
				while(k++ < i) ii++;
				return ii->second;
			}

			virtual unsigned int getNumNeighbors(unsigned int j) const{
				return neighbors->at(j)->size();
			}

			virtual unsigned int getNumNeighborCounts() const{
				return neighborCounts.size();
			}

			virtual unsigned int getNodeWithNeighborCount(unsigned int j, unsigned int i) const{
				auto jj = neighborCounts.begin();
				int h = 0;
				while(h++ < j) jj++;

				auto ii = (jj->second).begin();
				int k = 0;
				while(k++ < i) ii++;
				return *ii;
			}

			virtual unsigned int getNumNeighborSize(unsigned int j) const{
				auto ii = neighborCounts.begin();
				int k = 0;
				while(k++ < j) ii++;
				return ii->first;
			}

			virtual unsigned int getNumNeighborCounts(unsigned int j) const{
				auto ii = neighborCounts.begin();
				int k = 0;
				while(k++ < j) ii++;
				return (ii->second).size();
			}

			friend std::istream& operator>> (std::istream& in, MRF::Graph& g){
				g.loadGraph(in);
				g.init();
				return in;
			}

			friend std::ostream& operator<< (std::ostream& out, const MRF::Graph& g){
				g.storeGraph(out);
				return out;
			}
	};
};
#endif
