// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __MRF_HPP__
#define __MRF_HPP__

#include <rfdk/AbstractMRF.hpp>

namespace MRF{

	template <class R> class MRF : public AbstractMRF<R>{

		protected:
			using AbstractMRF<R>::initMsg;
			using AbstractMRF<R>::out_msg;
			using AbstractMRF<R>::in_msg;
			using AbstractMRF<R>::probs;
			using AbstractMRF<R>::dim;
			using AbstractMRF<R>::X;
			using AbstractMRF<R>::G;

//			using AbstractMRF<R>::marginals;

			R stepsize;
			R eps;
			R stddev;

			using AbstractMRF<R>::alpha;

			void postMessage(unsigned int i, bool reverse){
				R _z = 0;
				unsigned int o = 0;
				if(reverse)o=G->getNumEdges()*X;
				for(unsigned int k=0; k<X; ++k){
					_z += exp(out_msg[o+i*X+k]);
				}
				for(unsigned int k=0; k<X; ++k){
					const unsigned int idx = o+i*X+k;
					R t = in_msg[idx];
					R q = out_msg[idx] - std::log(_z);
					in_msg[idx] = log(alpha*exp(t) + (1-alpha)*exp(q));
					//in_msg[idx] = q;
					assert(t==t);
					out_msg[idx] = t;
				}
			}

			inline __attribute__((always_inline)) R accumulateProd(R a, R b) const {
				return a+b;
			}

			inline __attribute__((always_inline)) R potential(R a, R b) const {
				return exp(a+b);
			}

			inline __attribute__((always_inline)) R result(R a) const {
				return std::log(a);
			}

			inline __attribute__((always_inline)) double presult(R a, R b) const {
				return a/b;
			}

			virtual void storeConfig(std::ostream& out) const {
				out << X << std::endl;
				out << stepsize << std::endl;
				out << eps << std::endl;
				out << stddev << std::endl;
			}

			virtual void loadConfig(std::istream& in){
				in >> X;
				in >> stepsize;
				in >> eps;
				in >> stddev;
			}

		public:
			using AbstractMRF<R>::getProbs;
			using AbstractMRF<R>::marginals;

			MRF(R sd = 1.0) : stddev(sd), AbstractMRF<R>(){
				// can only be initialized from file or derived class
			}

			MRF(Graph* _G, unsigned int _numStates, R sd = 1.0, bool createParameterStorage = true) : stddev(sd), AbstractMRF<R>(_G, _numStates, createParameterStorage){
				initMsg = 0;

				//randomWeights();

				stepsize = 0.1;
				eps = 0.0001;
			}

			~MRF(){
				if(AbstractMRF<R>::par){
					//std::cout << "DELETING ((ParameterStorage<R>*)AbstractMRF::par) at " << MRF::AbstractMRF::par << std::endl;
					delete ((ParameterStorage<R>*)AbstractMRF<R>::par);
				}
			}

			void randomWeights(){
				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				for(unsigned int i=0; i<n; ++i){
					for(unsigned int j=0; j<X; ++j){
							const unsigned int idx = i*X+j;
							this->setWeight(idx, stddev * rand_gauss());
					}
					const unsigned int rid = i*X + (rand()%X);
					R val = this->getWeight(rid);
					this->setWeight(rid, val); // +4.0
				}
				for(unsigned int i=0; i<m; ++i){
					for(unsigned int j=0; j<X; ++j){
						for(unsigned int k=0; k<X; ++k){
							const unsigned int idx = n*X + i*X*X + j*X + k;
							this->setWeight(idx, stddev * rand_gauss());
						}
					}
				}
			}

			void setStepsize(R s){
				stepsize = s;
			}

			void setEps(R b){
				eps = b;
			}

			R getEps(){
				return eps;
			}

			R estimate(R* F, unsigned int N){

				const unsigned int n = G->getNumNodes();
				const unsigned int m = G->getNumEdges();

				R obj, r;

				R* empirical = new R[dim];

				for(unsigned int i=0; i<dim; ++i) empirical[i] = F[i]/(R)N;

				for(unsigned int i=0; i<dim; ++i){
					if(i<n*X){
						this->setWeight(i, std::log(empirical[i]));
					}else{
						const unsigned int eid = (i-n*X)/(X*X);
						const unsigned int t = ((i-n*X)-(eid*X*X));
						const unsigned int x = t/X;
						const unsigned int y = t%X;

						this->setWeight(i, std::log(empirical[i]) - std::log(empirical[G->getEdgeSource(eid)*X+x]) - std::log(empirical[G->getEdgeTarget(eid)*X+y]));
//						weights[i] = std::log(empirical[i]/(empirical[G->getEdgeSource(eid)*X+x]*empirical[G->getEdgeTarget(eid)*X+y]))/std::log(2);
					}
				}

				marginals();
				obj = -AbstractMRF<R>::logLikelihood(F, N);
				r   = l2diff(AbstractMRF<R>::getProbs(), empirical, dim);
				if(VERBOSE) std::cout << '\r' << 1 << "\t-l(w) = " << obj
								<< " |w|_1 = " << l1norm(AbstractMRF<R>::getWeights(), dim)
								<< " |g|_2 = " << r
								<< std::endl;

				delete empirical;

				return obj;
			}

			R optimize(R* F, unsigned int N){
				R lobj, obj = std::numeric_limits<R>::max(), r = std::numeric_limits<R>::max();
				unsigned int iter = 0;

				R* g = new R[dim], *h = new R[dim];
				for(unsigned int i=0; i<dim; ++i){
					g[i] = R(0);
					h[i] = R(1);
				}

				R* empirical = new R[dim];
				for(unsigned int i=0; i<dim; ++i) empirical[i] = F[i]/(R)N;

				while(l2diff(g,h,dim) > eps){
					marginals();

					lobj = obj;
					obj = -AbstractMRF<R>::logLikelihood(F, N);
					r   = l2diff(AbstractMRF<R>::getProbs(), empirical, dim);

					if(obj > lobj){
#pragma omp parallel for
						for(unsigned int i=0; i<dim; ++i){
							R oval = this->getWeight(i);
							R val = oval - stepsize * h[i];
							this->setWeight(i, val);
						}

						stepsize = stepsize/2;
						obj = lobj;
					}

					if(VERBOSE) std::cout << '\r' << std::fixed << std::setprecision(4) << ++iter << "    -l(w) = " << obj << "    |g|_2 = " << r << "   s = " << stepsize << std::flush;

#pragma omp parallel for
					for(unsigned int i=0; i<dim; ++i){
						g[i] = h[i];
						h[i] = empirical[i] - probs[i];

						R oval = this->getWeight(i);
						R val = oval + stepsize * h[i];
						this->setWeight(i, val);

						//assert(val == this->getWeight(i));
					}
				}

				marginals();
				obj = -AbstractMRF<R>::logLikelihood(F, N);
				r   = l2diff(AbstractMRF<R>::getProbs(), empirical, dim);
				if(VERBOSE)  std::cout << '\r' << std::fixed << std::setprecision(4)  << ++iter << "    -l(w) = " << obj << "    |g|_2 = " << r << " " << std::endl;

				delete[] g;
				delete[] h;
				delete[] empirical;

				return obj;
			}
	};
};
#endif
