// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __UTILS_HPP__
#define __UTILS_HPP__

#include <cstdlib>
#include <cmath>
#include <vector>
#include <sys/time.h>


namespace MRF{

	size_t SEED;
	unsigned int VERBOSE;
	unsigned int VERBOSEBP;
	unsigned int VERBOSEGRAPH;

	typedef std::pair<unsigned int, unsigned int> _pair;
	typedef std::vector<_pair* > _vec;
	typedef std::pair<unsigned int, std::pair<unsigned int, unsigned int> > _spair;

	template <typename T> int sgn(T val){
		return (T(0) < val) - (val < T(0));
	}

	template <class T> T get(char* in){
		std::stringstream sstr(in);
		T a;
		sstr >> a;
		return a;
	}

	template <typename T, typename R>
	bool compFirst(std::pair<T, R> a, std::pair<T, R> b){
		return (a.first > b.first);
	}

	double str2dbl(std::string s){
		std::stringstream is;
		is << s;
		double r;
		is >> r;
		return r;
	}

	int str2int(std::string s){
		std::stringstream is;
		is << s;
		int r;
		is >> r;
		return r;
	}

	unsigned int str2uint(std::string s){
		std::stringstream is;
		is << s;
		unsigned int r;
		is >> r;
		return r;
	}

	double rand_gauss(){
		double v1, v2, s;

		do{
			v1 = 2.0 * ((double) rand()/RAND_MAX) - 1;
			v2 = 2.0 * ((double) rand()/RAND_MAX) - 1;

			s = v1*v1 + v2*v2;
	  	}while(s >= 1.0);

		  	if(s == 0.0) return 0.0;
		else return (v1*sqrt(-2.0 * std::log(s)/s));
	}

	// Upper bound on log_2
	inline __attribute__((always_inline)) size_t bl(size_t x){
		return (x ? (sizeof(size_t)*8)-__builtin_clzl(x) : x);
	}

	// Lower bound on log_2
	inline __attribute__((always_inline)) size_t log2lb(size_t x){
		size_t b = bl(x);
		if(x) return b-1;
		return x;
	}

	template <typename T> T l1diff(T* a, T* b, size_t n){
		T r = 0;
		for(int i=0; i<n; ++i){
			const long val = long(a[i]) - b[i];
			//std::cout << val << " = " << a[i] << " - " << b[i] << std::endl;
			r += std::abs(val);
		}
		return r;
	}

	template <typename T> T l1norm(T* a, size_t n){
		T r = 0;
		for(int i=0; i<n; ++i){
			r += std::abs(a[i]);
		}
		return r;
	}

	template <typename T> T l2norm(T* a, size_t n){
		T r = 0;
		for(int i=0; i<n; ++i){
			r += a[i]*a[i];
		}
		return std::sqrt(r);
	}

	template <typename T, typename R> T l2diff(T* a, R* b, const size_t n){
		T r = 0;
		for(int i=0; i<n; ++i){
			T x = std::max(a[i],T(b[i]));
			T y = std::min(a[i],T(b[i]));
			T val = 0L + x - y;
			r += val*val;
					if(r != r){
						asm("int3");
						//assert(false);
					}
		}
		return std::sqrt(r);
	}

	template <typename T, typename R> double dot(T* a, R* b, size_t n){
		double r = 0;
		for(int i=0; i<n; ++i){
			r += a[i] * b[i];
		}
		return r;
	}

	inline double getTime(){
		struct timeval tv;
		if (gettimeofday(&tv, NULL)==0)
			return tv.tv_sec+((double)(tv.tv_usec))/1e6;
		else
			return 0.0;
	}

	class Stat{
		private:
			double x;
			double xx;
			double l;
			size_t n;

		public:
			Stat(): n(0), x(0.0), xx(0.0), l(0.0) {};

			void add(double v){
				l = v;
				x  += v;
				xx += v*v;
				n++;
			}

			void undo(){
				if(n){
					x  -= l;
					xx -= l*l;
					n--;
				}
			}

			void reset(){
				n  = 0;
				x  = 0.0;
				xx = 0.0;
				l  = 0.0;
			}

			double mean(){
				return x/n;
			}

			double dev(){
				return std::sqrt( (xx - (x*x)/n)/(n-1) );
			}
	};
};
#endif
