// Copyright (c) 2013, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __ST_GRAPH_HPP__
#define __ST_GRAPH_HPP__

#include <cstring>

#include <rfdk/Graph.hpp>
#include <rfdk/utils.hpp>

namespace MRF{
	class STGraph : public Graph{ // this constructs a spatio-temporal graph as described in the paper, based on a given spatial graph G_0
		private:
			using Graph::neighbors;
			using Graph::n;
			using Graph::m;
			using Graph::edges;
			using Graph::keepEdges;
			using Graph::tree;
			using Graph::colors;


			using Graph::nN;
			using Graph::nNR;

			unsigned int T;
			unsigned int m0;
			unsigned int n0;

			std::map<_pair,unsigned int> edgesMap;

			void initGraph(){
				m = edges->size();

				_vec* e = new _vec();

				for(unsigned int i=0; i<m; ++i){
					edgesMap[*edges->at(i)] = edgesMap.size();
					//n = std::max(n, 1+std::max(edges->at(i)->first, edges->at(i)->second));
				}
				n0 = n;
				m0 = m;

				//std::cout << "STGraph: " << n0 << " SPATIAL NODES AND " << m0 << " SPATIAL EDGES" << std::endl;

				neighbors = new std::vector<std::set<_pair>*>();
				for(unsigned int i=0; i<n; ++i){
					for(unsigned int t = 0; t<T; ++t){
						neighbors->emplace_back(new std::set<_pair>()); // vertex id is i*T+t
						// corresponding weight id (i,t,x) is i*T*X+t*X+x
					}
				}
				// total number of vertices is n0*T
				//std::cout << "STGraph: " << n0*T << " SPATIO-TEMPORAL NODES" << std::endl;

				for(unsigned int i=0; i<n; ++i){
					for(unsigned int t = 0; t<T-1; ++t){
						unsigned int u = i*T+t;
						unsigned int v = i*T+t+1;
						neighbors->at(u)->insert(_pair(v, e->size()));
						neighbors->at(v)->insert(_pair(u, e->size()));
						e->emplace_back(new _pair(u, v)); // edge id is i*(T-1)+t
						// corresponding edge weight id (i,t,x,y) is n*T*X + i*(T-1)*X^2 + t*X^2 + x*X + y
					}
				}

				for(unsigned int i=0; i<m; ++i){
					//if(i%25==0) std::cout << i << std::endl;
					for(unsigned int t = 0; t<T; ++t){
						unsigned int u = edges->at(i)->first*T+t;
						unsigned int v = edges->at(i)->second*T+t;
						neighbors->at(u)->insert(_pair(v, e->size()));
						neighbors->at(v)->insert(_pair(u, e->size()));
						e->emplace_back(new _pair(u, v)); // edge id is n*(T-1) + i*((T-1)*3+1) + t*3 + 0
						// corresponding edge weight id (i,t,x,y) is n*T*X + n*(T-1)*X^2 + i*((T-1)*3+1)*X^2 + t*3*X^2 + x*X + y

						if(t<T-1){
							neighbors->at(u)->insert(_pair(v+1, e->size()));
							neighbors->at(v+1)->insert(_pair(u, e->size()));
							e->emplace_back(new _pair(u, v+1)); // edge id is n*(T-1) + i*((T-1)*3+1) + t*3 + 1
							// corresponding edge weight id (i,t,x,y) is n*T*X + n*(T-1)*X^2 + i*((T-1)*3+1)*X^2 + t*3*X^2 + X^2 + x*X + y

							neighbors->at(u+1)->insert(_pair(v, e->size()));
							neighbors->at(v)->insert(_pair(u+1, e->size()));
							e->emplace_back(new _pair(u+1, v)); // edge id is n*(T-1) + i*((T-1)*3+1) + t*3 + 2
							// corresponding edge weight id (i,t,x,y) is n*T*X + n*(T-1)*X^2 + i*((T-1)*3+1)*X^2 + t*3*X^2 + 2*X^2 + x*X + y
						}
					}
				}

				m = e->size(); // total number of edges is n*(T-1) + m*((T-1)*3+1)

				for(int i = 0; i<edges->size(); ++i){
//					std::cout << i << " deleting edge at " << edges->at(i) << std::endl;
					delete edges->at(i);
				}
				delete edges;
				edges = e;

				n*=T;
			}

			void init(){
				//n = 0;
				//m = 0;

				initGraph();
				//colors = new unsigned int[n];
				//computeColoring();
			}

		public:

			STGraph(std::istream& _G, unsigned int _T) : Graph() {

				loadGraph(_G);
				T = _T;
				init();
				tree = false;
				//std::cout << "STGraph(" << &_G << ", " << _T << ")" << std::endl;
			}

			STGraph(Graph* _G, unsigned int _T) : Graph() {

				_vec* E = _G->getEdges();
				edges = new _vec();

				for(auto e : *E){
					edges->push_back(new _pair(*e));
				}

				n = _G->getNumNodes();

				for(unsigned int i=0; i<n; ++i){
					std::string name = _G->getVertexNameById(i);
					nNR[name] = i+1;
					nN[i+1] = name;
				}

				T = _T;
				init();
				tree = false;
				//std::cout << "STGraph(" << &_G << ", " << _T << ")" << std::endl;
			}

			~STGraph(){

			}

			unsigned int getSpatialNodes(){
				return n0;
			}

			unsigned int getSpatialEdges(){ // rename to getNum...
				return m0;
			}

			unsigned int getNumLayers(){
				return T;
			}

			int getEdgeId(long u, long v){
				std::pair<long,long> e(u,v);
				if(edgesMap.find(e) != edgesMap.end()){
					return edgesMap[e];
				}
				return -1;
			}
	};
};
#endif
