// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __STRF_HPP__
#define __STRF_HPP__

#include <math.h>

#include <rfdk/MRF.hpp>
#include <strf/STGraph.hpp>
#include <strf/SparseSTParameter.hpp>

namespace MRF{

	template <class R> inline R soft(R z, R a){
		return sgn(z)*std::max(R(0), std::abs(z)-a);
	}

	template <class R> class STRF : public MRF<R>{

		protected:

			using AbstractMRF<R>::par;

			using MRF<R>::X;
			using MRF<R>::dim;
			using MRF<R>::probs;
			using MRF<R>::eps;

			R lam1;
			R lam2;

			R Dmin;
			R Dmax;

			R lower;

			unsigned int T;
			unsigned int mI;

			bool useTheta;
			bool useSpatialEdgeW;
			bool useVertexW;
		public:
			using AbstractMRF<R>::getProbs;
			using AbstractMRF<R>::marginals;
			using AbstractMRF<R>::MAP;

			STRF(R sd = 1.0) : MRF<R>(sd){
				useTheta = false;
			}

			STRF(STGraph* _G, unsigned int _numStates, const unsigned int _T) : MRF<R>(_G, _numStates, 1.0, false){
				T = _T;

				useSpatialEdgeW = true;
				useVertexW = true;

				lam1 = 0.001; // auto tune me!
				lam2 = 0.001; // auto tune me!

				Dmin = 0.1; // auto tune me!
				Dmax = 0.1; // auto tune me!

				mI = -1;

				lower = 0.0000001; // auto tune me!

				par = new SparseSTParameterStorage<R>((STGraph*)_G, _numStates, _T);
				//std::cout << "SETTING (SparseSTParameterStorage<R>*)AbstractMRF::par TO " << par << std::endl;

				useTheta = false;
			}

			~STRF(){
				if(par){
					//std::cout << "DELETING ((SparseSTParameterStorage<R>*)AbstractMRF::par) at " << par << std::endl;
					delete ((SparseSTParameterStorage<R>*)par);
					par = 0;
				}
			}

			void useSpatialEdgeWeights(bool b){
				useSpatialEdgeW = b;
			}

			void useVertexWeights(bool b){
				useVertexW = b;
			}

			void setMaxIter(unsigned int v){
				mI = v;
			}

			void setLambda1(R v){
				lam1 = v;
			}

			void setLambda2(R v){
				lam2 = v;
			}

			void setDinvmin(R v){
				Dmin = v;
			}

			void setDinvmax(R v){
				Dmax = v;
			}

			void setDesiredIncrement(R v){
				lower = v;
			}

			R getLambda1(){
				return lam1;
			}

			R getLambda2(){
				return lam2;
			}

			R getDinvmin(){
				return Dmin;
			}

			R getDinvmax(){
				return Dmax;
			}

			R getDesiredIncrement(){
				return lower;
			}

			void setObservation(unsigned int v, unsigned int t, int x){
				AbstractMRF<R>::setObservation(v*T+t, x);
			}

			int getObservation(unsigned int v, unsigned int t){
				return AbstractMRF<R>::getObservation(v*T+t);
			}

			unsigned int getMAPState(unsigned int v, unsigned int t){
				return AbstractMRF<R>::getMAPState(v*T+t);
			}

			double getProb(unsigned int v, unsigned int t, unsigned int x){
				return AbstractMRF<R>::getProb(v*T+t,x);
			}

			R getLogPotential(const unsigned int i) const {
				if(!useTheta) return ((SparseSTParameterStorage<R>*)par)->getSum(i);
				return ((SparseSTParameterStorage<R>*)par)->get(i);
			}

			void useCompressedParameters(bool b){
				useTheta = !b;
			}

			void shrink(){
				((SparseSTParameterStorage<R>*)par)->shrink();
			}

			unsigned int nnz(){
				return ((SparseSTParameterStorage<R>*)par)->nnz();
			}

			unsigned int getNumSpatialNodes(){
				return ((SparseSTParameterStorage<R>*)par)->getGraph()->getSpatialNodes();
			}

			unsigned int getNumSpatialEdges(){
				return ((SparseSTParameterStorage<R>*)par)->getGraph()->getSpatialEdges();
			}

			R optimize(R* F, unsigned int N){

				R obj, r = std::numeric_limits<R>::max();
				unsigned int iter = 0;

				SparseSTParameterStorage<R>* p = (SparseSTParameterStorage<R>*)par;
				SparseSTParameterStorage<R>* old = 0;

				unsigned int n = p->getGraph()->getSpatialNodes();
				unsigned int m = p->getGraph()->getSpatialEdges();

				R* e = new R[dim];
#pragma omp parallel for
				for(unsigned int i=0; i<dim; ++i) e[i] = F[i]/(R)N;

				R* h = new R[dim];
				R* g = new R[dim];
#pragma omp parallel for
				for(unsigned int i=0; i<dim; ++i){
					h[i] = R(10);
					g[i] = R(0);
				}

				// Estimated curvatures for each spatio-temporal component of the graph
				R* Dinv = new R[2*n+3*m];
#pragma omp parallel for
				for(unsigned int i=0; i<2*n+3*m; ++i){
					Dinv[i] = Dmax;
				}

				marginals();
				obj = -AbstractMRF<R>::logLikelihood(F, N);

				while(l2diff(g,h,dim) > eps && iter < mI){
					memcpy(g, h, dim*sizeof(R));

					// compute partial derivaties of smooth part:
#pragma omp parallel for
					for(unsigned int i=0; i<dim; ++i){

						int x = -1, y = -1, v = -1, u = -1, t = -1, k = -1;
						p->decode(i, x, y, v, u, t, k);

						// T-1 steps for temporal and spatio-temporal edges (T steps for nodes and spatial edges):
						unsigned int _T = T;
						if((u==-1 && k>0) || (u!=-1 && k==-1)) _T--;

						R temp = 0;
						for(long j=t; j<_T; ++j){
							unsigned int IDX = p->encode(x,y,v,u,j,k);
							temp += p->getCoefficient(j-t)*(e[IDX]-probs[IDX]);
						}
						R val = p->get(i);
						h[i] = -temp + lam2*val;
					}

					// estimate curvatures
#pragma omp parallel for
					for(int c=0; c<n+m; ++c){
						for(int k=0; (k<1 && c<n) || (k<3 && c>=n); ++k){
							R JJv = 0, JZv = 0;
							R JJt = 0, JZt = 0;
							R JJs = 0, JZs = 0;
							for(int x=0; x<X; ++x){
								for(int y=0; y<X; ++y){
									for(int t=0; t<T; ++t){
										// vertex or temporal-edge
										if(c<n){
											if(y==0){

												// vertex
												unsigned int idx = p->encode(x,-1,c,-1,t,-1);
												if(old){
													R J = h[idx] - g[idx];
													R Z = p->get(idx) - old->get(idx);

													JJv += J*J;
													JZv += J*Z;
												}

											}

											if(t==T-1) break;
											// temporal edge
											unsigned int idx = p->encode(x,y,c,c,t,-1);

											if(old){
												R J = h[idx] - g[idx];
												R Z = p->get(idx) - old->get(idx);

												JJt += J*J;
												JZt += J*Z;
											}

										// spatio(-temporal) edge
										}else{
											if(t==T-1 && k>0) break;
											// spatio-temporal edge
											unsigned int idx = p->encode(x,y,c-n,-1,t,k);
											if(old){
												R J = h[idx] - g[idx];
												R Z = p->get(idx) - old->get(idx);

												JJs += J*J;
												JZs += J*Z;
											}
										}
									}
								}
							}
							if(old) if(c<n){
								if(JJv != 0) Dinv[c] = JZv/JJv;
								Dinv[c] = std::min(std::max(Dinv[c], Dmin), Dmax);

								if(JJt != 0) Dinv[n+c] = JZt/JJt;
								Dinv[n+c] = std::min(std::max(Dinv[n+c], Dmin), Dmax);
							}
							if(old) if(c>=n){
								const unsigned int idx = 2*n+k*m+(c-n);
								if(JJs != 0) Dinv[idx] = JZs/JJs;
								Dinv[idx] = std::min(std::max(Dinv[idx], Dmin), Dmax);
							}
						}
					}

					const unsigned int max = 1024;
					// line search
#pragma omp parallel for
					for(int c=0; c<n+m; ++c){
						for(int k=0; (k<1 && c<n) || (k<3 && c>=n); ++k){
							R fv = 0;
							R ft = 0;
							R fs = 0;
							bool done_a = false, done_b = true;
							unsigned int ii = 0;
							do{
								for(int t=0; t<T; ++t){
									for(int x=0; x<X; ++x){
										for(int y=0; y<X; ++y){
											// vertex or temporal-edge
											if(c<n){
												if(y==0){
													// vertex
													unsigned int idx = p->encode(x,-1,c,-1,t,-1);
													if(old){
														R a = p->get(idx) - h[idx]*Dinv[c];
														R b = lam1*Dinv[c];
														R newval = soft(a, b);

														fv += h[idx]*( newval - p->get(idx) ) + lam1*( std::abs(newval) - std::abs(p->get(idx)) );
													}
												}
												if(t==T-1) break;
												// temporal edge
												unsigned int idx = p->encode(x,y,c,c,t,-1);
												if(old){
													R a = p->get(idx) - h[idx]*Dinv[n+c];
													R b = lam1*Dinv[n+c];
													R newval = soft(a, b);

													ft += h[idx]*( newval - p->get(idx) ) + lam1*( std::abs(newval) - std::abs(p->get(idx)) );
												}
											// spatio(-temporal) edge
											}else{
												if(t==T-1 && k>0) break;
												// spatio-temporal edge
												unsigned int idx = p->encode(x,y,c-n,-1,t,k);
												if(old){
													int j = 2*n+k*m+(c-n);
													R a = p->get(idx) - h[idx]*Dinv[j];
													R b = lam1*Dinv[j];
													R newval = soft(a, b);

													fs += h[idx]*( newval - p->get(idx) ) + lam1*( std::abs(newval) - std::abs(p->get(idx)) );
												}
											}
										}
									}
								}
								R f;
								unsigned int j;
								if(c<n){
									j = c;
									f = fv;
									done_b = false;

									const unsigned int idx = n+c;
									if(ft > -lower){
										//std::cout << "ft = " << ft << std::endl;
										R d = Dinv[idx]*0.5;
										if(d<Dmin){
											done_b = true;
											continue;
										}else Dinv[idx] = d;
									}else done_b = true;

								}else if(c>=n){
									j = 2*n+k*m+(c-n);
									f = fs;
									done_b = true;
								}
								if(f > -lower){
									//std::cout << "fv = " << fv << std::endl;
									R d = Dinv[j]*0.5;
									if(d<Dmin){
										done_a = true;
										continue;
									}else Dinv[j] = d;
								}else done_a = true;
							}while(!(done_a && done_b) && ++ii<max);
						}
					}

					// store weights for next iteration
					if(old) delete old;
					old = new SparseSTParameterStorage<R>(*p);

					R eta = 0.0, eta2 = 0.0;
#pragma omp parallel for
					for(unsigned int i=0; i<dim; ++i){
						int x = -1, y = -1, c = -1, u = -1, t = -1, k = -1;
						p->decode(i, x, y, c, u, t, k);

						unsigned int idx = c;
						if(c<n && c==u) idx = n+c;
						else if(c>=n) idx = 2*n + k*m + (c-n);

						R a = p->get(i) - h[i]*Dinv[idx];
						R b = lam1*Dinv[idx];
						if((useVertexW || i>=n*T*X) && (useSpatialEdgeW || k!=0)){
							h[i] -= lam2*(p->get(i));
							eta += Dinv[idx];
							eta2 += Dinv[idx]*Dinv[idx];

							p->set(i, soft(a, b)); // SpaRSA
						}
					}

					marginals();

					R* W = p->getArray();

					R L1 = l1norm(W, dim);
					R L2 = l2norm(W, dim);

					obj = -AbstractMRF<R>::logLikelihood(F, N) + lam1*L1 + 0.5*lam2*L2*L2;
					r = 0;

					for(unsigned int i=0; i<dim; ++i){
						int x = -1, y = -1, c = -1, u = -1, t = -1, k = -1;
						p->decode(i, x, y, c, u, t, k);
						if((useVertexW || i>=n*T*X) && (useSpatialEdgeW || k!=0)) r += std::abs(h[i]);
					}

					R gvar = 0.0;
					for(unsigned int i=0; i<dim; ++i){
						gvar += h[i]*h[i];
					}

					gvar = sqrt(gvar - (r*r));
					R etavar = sqrt(eta2/dim - (eta/dim)*(eta/dim));

					std::cout << std::fixed << std::setprecision(4) << '\r' << 
						"iter = " << ++iter << 
						" obj = " << obj << 
						"\t|grad|_1 = " << r  << 
//						"  |Z|_0/d = " << nnz()/R(dim) << 
						"\tstep = " << (eta/dim) <<"+-"<< etavar << std::flush;
					std::cout << std::endl;
				}
				std::cout << std::endl;

				delete h;
				delete g;
				delete e;
				delete Dinv;

				return obj;
			}
	};
};
#endif
