// Copyright (c) 2013, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __SPARAMETER_HPP__
#define __SPARAMETER_HPP__

#include <iostream>
#include <iomanip>
#include <algorithm>
#include <cmath>
#include <limits>
#include <cassert>
#include <cstring>
#include <map>

#include <rfdk/Parameter.hpp>
#include <strf/STGraph.hpp>

#ifndef DEBUG
#define DEBUG false
#endif

namespace MRF{

	template <class R> class SparseSTParameterStorage : public AbstractParameterStorage<R>{

		private:

			template <class S, class T> class VectorMap{ // use this instead of a "map" if its size is "small const."
				public:
					std::vector<S> idx;
					std::vector<T> val;

					VectorMap(){

					}


					~VectorMap(){

					}

					unsigned int size(){
						return idx.size();
					}

					void add(S i, T v){

						auto ii = idx.begin();
						auto jj = val.begin();

						for(; ii!=idx.end(); ++ii){
							if(*ii == i){
								if(v != 0) *jj = v;
								else{
									//std::cout << "DELETE " << i << "   " << *ii << std::endl;
									ii = idx.erase(ii);
									jj = val.erase(jj);
									//std::cout << "DELETE POST, size = " << idx.size() << std::endl;
								}
								return;
							}else{
								++jj;
							}
						}
						//std::cout << "ADDING " << i << "   " << v << std::endl;
						if(v != 0){
							idx.push_back(i);
							val.push_back(v);
						}
					}

					T get(const S i) const {
						for(register unsigned int j=0; j<idx.size(); ++j) if(idx[j] == i){
							return val[j];
						}

						return 0;
					}

					T sum(const S t, T* coeff) const {
						const unsigned int _n = idx.size();

						register T s = 0.0;

						if(_n) for(register unsigned int j=0, i=idx[j]; j<_n && i<=t; ++j){
							i = idx[j];
							if(i>t) break;
							T v = val[j];
							T cv = coeff[t-i];
							s += v*cv;
						}

						return s;
					}

					void shrink(){
						idx.shrink_to_fit();
						val.shrink_to_fit();
					}
			};

			unsigned int _n;
			unsigned int m;
			unsigned int X;
			unsigned int T;

			R* coeff;

			R* data;
			using AbstractParameterStorage<R>::n;

			bool arrayReady;

			STGraph* G;
			std::map<unsigned int,VectorMap<unsigned int,R> >*** jump;

		public:

			SparseSTParameterStorage(SparseSTParameterStorage<R>& other) : AbstractParameterStorage<R>(other.getGraph()->getNumNodes()*other.getDom() + other.getGraph()->getNumEdges()*other.getDom()*other.getDom()){

				G = other.getGraph();
				X = other.getDom();
				T = other.getLayers();

				_n = G->getSpatialNodes();
				m = G->getSpatialEdges();

				data = 0;
				arrayReady = false;

				coeff = new R[T];
				memcpy(coeff, other.getCoefficients(), T*sizeof(R));

				std::map<unsigned int,VectorMap<unsigned int,R> >*** ojump = other.getJumptable();

				jump = new std::map<unsigned int,VectorMap<unsigned int,R> >**[_n+m];

				for(unsigned int i=0; i<_n+m; ++i){
					if(ojump[i]){
						jump[i] = new std::map<unsigned int,VectorMap<unsigned int,R> >*[3];

						for(unsigned int j=0; j<3; ++j){
							if(ojump[i][j]){
								jump[i][j] = new std::map<unsigned int,VectorMap<unsigned int,R> >();

								for(auto ii = ojump[i][j]->begin(); ii != ojump[i][j]->end(); ++ii){
									(*jump[i][j])[ii->first] = VectorMap<unsigned int,R>(ii->second);
								}
								
							}else{
								jump[i][j] = 0;
							}
						}
						
					}else{
						jump[i] = 0;
					}
				}
			}

			SparseSTParameterStorage(STGraph* _G, const unsigned int _X, const unsigned int _T) : AbstractParameterStorage<R>(_G->getNumNodes()*_X + _G->getNumEdges()*_X*_X){

				G = _G;
				X = _X;
				T = _T;

				data = 0;
				arrayReady = false;

				_n = G->getSpatialNodes();
				m = G->getSpatialEdges();

				coeff = new R[T];
				coeff[0] = 1;

				for(unsigned int t=0; t<T; ++t){
					for(unsigned int i=0; i<t; ++i){
						coeff[t-i] = 1.0/(t-i+1); // precompute temporal weights
					}
				}

				jump = new std::map<unsigned int,VectorMap<unsigned int,R> >**[_n+m];
				for(unsigned int i=0; i<_n+m; ++i){
					jump[i] = 0;
				}
			}

			std::map<unsigned int,VectorMap<unsigned int,R> >*** getJumptable(){
				return jump;
			}

			STGraph* getGraph(){
				return G;
			}

			unsigned int getDom(){
				return X;
			}

			unsigned int getLayers(){
				return T;
			}

			~SparseSTParameterStorage(){
				for(unsigned int i=0; i<_n+m; ++i){
					if(jump[i]){
						for(unsigned int j=0; j<3; ++j){
							if(jump[i][j]) delete jump[i][j];
						}
						delete[] jump[i];
					}
				}
				if(jump) delete[] jump;
				//std::cout << "DELETING SparseSTParameterStorage::coeff at " << coeff << std::endl;
				if(coeff) delete[] coeff;
				if(data) delete[] data;
			}

			R getCoefficient(unsigned int j){
				return coeff[j];
			}

			R* getCoefficients(){
				return coeff;
			}

			R* getArray(){
				if(!(data)) data = new R[n];
				if(!arrayReady){
					for(unsigned int i=0; i<n; ++i) data[i] = getSum(i);
					arrayReady = true;
				}
				return data;
			}

			void zero(){
				for(unsigned int i=0; i<_n+m; ++i){
					if(jump[i]){
						for(unsigned int j=0; j<3; ++j){
							if(jump[i][j]) delete jump[i][j];
							jump[i][j] = 0;
						}
						delete[] jump[i];
						jump[i] = 0;
					}
				}
			}

			R get(int x, int y, int v, int u, int t, int edgeType) const {

				if(x>=0 && v>=0 && t>=0){
					if(y == -1){ // vertex
						if(jump[v]){
							if(jump[v][0]){
								std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[v][0];
								VectorMap<unsigned int,R>& MX = (*M)[x];
								R val = MX.get(t);

								return val;
							}
						}
						return 0;

					}else if(v==u){ // temporal edge
						if(jump[v]){
							if(jump[v][1]){
								std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[v][1];
								VectorMap<unsigned int,R>& MX = (*M)[x*X+y];
								R val = MX.get(t);

								return val;
							}
						}
						return 0;

					}else if(edgeType>=0 && edgeType<3){ // spatio(-temporal) edge

						int eid = v;

						if(jump[_n+eid]){
							if(jump[_n+eid][edgeType]){
								std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[_n+eid][edgeType];
								VectorMap<unsigned int,R>& MX = (*M)[x*X+y];
								R val = MX.get(t);

								return val;
							}
						}
						return 0;
					}
				}
				std::cout << x << " " << v << " " << t << " " << y << " " << u << std::endl;
				assert(false);
				return 0;
			}

			R get(const size_t i) const {
				int x = -1, y = -1, v = -1, u = -1, t = -1, edgeType = -1;
				decode(i, x, y, v, u, t, edgeType);
				return get(x, y, v, u, t, edgeType);
			}

			unsigned int nnz(){
				unsigned int result = 0;

				for(unsigned int i=0; i<_n+m; ++i){
					if(jump[i]){
						for(unsigned int j=0; j<3; ++j){
							//if(jump[i][j]) std::cout << "jump[" << i << "][" << j << "]->size() = " << jump[i][j]->size() << std::endl;
							// result += jump[i][j]->size();
							if(jump[i][j]){
								for(auto ii = jump[i][j]->begin(); ii != jump[i][j]->end(); ++ii){
									//std::cout << i << "   "  << j << "   " << ii->first << "   " << ii->second.size() << std::endl;
									result += ii->second.size();
								}
							}
						}
					}
				}

				return result;
			}

			R getSum(int x, int y, int v, int u, int t, int edgeType) const {

				if(x>=0 && v>=0 && t>=0){
					if(y == -1){ // vertex
						if(jump[v]){
							if(jump[v][0]){
								std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[v][0];
								VectorMap<unsigned int,R>& MX = (*M)[x];

								//__builtin_prefetch((void *)MX.idx.data, 0 , 3);
								//__builtin_prefetch((void *)MX.val.data, 0 , 3);

								R sum = MX.sum(t, coeff);

								return sum;
							}
						}

					}else if(v==u){ // temporal edge
						if(jump[v]){
							if(jump[v][1]){
								std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[v][1];
								VectorMap<unsigned int,R>& MX = (*M)[x*X+y];

								//__builtin_prefetch((const void *)MX.idx.data, 0 , 3);
								//__builtin_prefetch((const void *)MX.val.data, 0 , 3);

								R sum = MX.sum(t, coeff);

								return sum;
							}
						}

					}else if(edgeType>=0 && edgeType<3){ // spatio(-temporal) edge

						int eid = v;

						if(jump[_n+eid]){
							if(jump[_n+eid][edgeType]){
								std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[_n+eid][edgeType];
								VectorMap<unsigned int,R>& MX = (*M)[x*X+y];

								//__builtin_prefetch((const void *)MX.idx.data, 0 , 3);
								//__builtin_prefetch((const void *)MX.val.data, 0 , 3);

								R sum = MX.sum(t, coeff);

								return sum;
							}
						}				
					}
				}

				return 0;
			}

			R getSum(const unsigned int i) const {
				int x = -1, y = -1, v = -1, u = -1, t = -1, edgeType = -1;
				decode(i, x, y, v, u, t, edgeType);
				getSum(x, y, v, u, t, edgeType);
			}

			void shrink(){
				for(unsigned int i=0; i<_n+m; ++i){
					if(jump[i]){
						for(unsigned int j=0; j<3; ++j){
							if(jump[i][j])
								for(auto ii = jump[i][j]->begin(); ii != jump[i][j]->end(); ++ii){
									ii->second.shrink();
								}
						}
					}
				}
			}

			void set(int x, int y, int v, int u, int t, int edgeType, const R val) {
				arrayReady = false;
				if(x>=0 && v>=0 && t>=0){
					if(y == -1){ // vertex

						if(!jump[v]){
							jump[v] = new std::map<unsigned int,VectorMap<unsigned int,R> >*[3];
							for(int z=0; z<3; ++z) jump[v][z] = 0;
						}

						if(!jump[v][0]){
							jump[v][0] = new std::map<unsigned int,VectorMap<unsigned int,R> >();
						}

						std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[v][0];
						VectorMap<unsigned int,R>& MX = (*M)[x];
						MX.add(t, val);

					}else if(v==u){
						if(!jump[v]){
							jump[v] = new std::map<unsigned int,VectorMap<unsigned int,R> >*[3];
							for(int z=0; z<3; ++z) jump[v][z] = 0;
						}

						if(!jump[v][1]){
							jump[v][1] = new std::map<unsigned int,VectorMap<unsigned int,R> >();
						}

						std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[v][1];
						VectorMap<unsigned int,R>& MX = (*M)[x*X+y];
						MX.add(t, val);

					}else if(edgeType>=0 && edgeType<3){ // spatio(-temporal) edge

						int eid = v;

						if(!jump[_n+eid]){
							jump[_n+eid] = new std::map<unsigned int,VectorMap<unsigned int,R> >*[3];
							for(int z=0; z<3; ++z) jump[_n+eid][z] = 0;
						}

						if(!jump[_n+eid][edgeType]){
							jump[_n+eid][edgeType] = new std::map<unsigned int,VectorMap<unsigned int,R> >();
						}

						std::map<unsigned int,VectorMap<unsigned int,R> >* M = jump[_n+eid][edgeType];
						VectorMap<unsigned int,R>& MX = (*M)[x*X+y];
						MX.add(t, val);

					}else{
						std::cerr << "ERROR1: CANNOT DECODE ID!" << std::endl;
					}
				}else{
					std::cerr << "ERROR0: CANNOT DECODE ID!" << std::endl;
				}
			}

			void set(const size_t i, const R val) {
				int x = -1, y = -1, v = -1, u = -1, t = -1, edgeType = -1;
				decode(i, x, y, v, u, t, edgeType);
				set(x, y, v, u, t, edgeType, val);
				//if(i>=_n*T*X && i<_n*T*X+20) std::cout << "w[" << i << "] = " << std::fixed << std::setprecision(16) << get(x, y, v, u, t, edgeType) << std::endl;
			}

			R& operator[](const unsigned int i){
				//return data[i];
				assert(false);
			}

			void decode(const unsigned int i, int& dx, int& dy, int& di, int& dj, int& dt, int& dk) const {

				int IDX;

				const int XX = X*X;
				const int nTX = _n*T*X;
				const int nTm1XX = _n*(T-1)*XX;

				const unsigned int dim = nTX+nTm1XX+m*(3*(T-1)+1)*XX;

				assert(i < dim);

				if(i < nTX){ // i = j*T*X + t*X + x
					dx = i % X; // = x
					int temp = (i-dx)/X; // = j*T + t
					dt = temp % T; // = t
					di = (temp-dt)/T; // = j

					IDX = di*T*X + dt*X + dx;
				}

				else if(i < nTX + nTm1XX){ // i = _n*T*X + j*(T-1)*(X*X) + t*(X*X) + x*X + y

					int temp = i-nTX; // = j*(T-1)*(X*X) + t*(X*X) + x*X + y
					dy = i % X; // = y

					dx = (temp-dy)%XX; // = x*X
					dx /= X; // = x

					dt = (temp-(dx*X+dy))/XX; // = j*(T-1) + t
					dt %= (T-1); // = t

					di = (((temp-(dx*X+dy))/XX)-dt)/(T-1); // = j
					dj = di;

					IDX = nTX + di*(T-1)*XX + dt*XX + dx*X + dy;
				}

				else{ // i = _n*T*X + _n*(T-1)*(X*X) + j*(3*(T-1)+1)*(X*X) + t*3*(X*X) + k*(X*X) + x*X + y

					int temp = i-(nTX+nTm1XX); // = j*(3*(T-1)+1)*(X*X) + t*3*(X*X) + k*(X*X) + x*X + y

					dy = i % X; // = y

					dx = (temp-dy)%XX; // = x*X
					dx /= X; // = x

					dt = (temp-(dx*X+dy))/XX; // = j*((T-1)*3+1) + t*3 + k
					dt %= ((T-1)*3+1); // if t==T-1 && k>0: dt = k-1, else: dt = t*3 + k
					dt /= 3; // = t + (int)(k/3) = t

					dk = (temp-(dx*X+dy))/XX; // = j*((T-1)*3+1) + t*3 + k
					dk %= ((T-1)*3+1); // = t*3 + k
					dk -= dt*3;

/*
					dk = (temp-(dx*X+dy))/XX; // = j*(3*(T-1)+1) + t*3 + k
					dk %= 3; // = k

					dt = (temp-(dx*X+dy+dk*XX))/XX; // = j*(3*(T-1)+1) + t*3
					dt %= (3*(T-1)+1); // = t*3
					dt /= 3;
*/

					di = (temp-(dx*X+dy+dk*XX+dt*3*XX))/(((T-1)*3+1)*XX); // = j
					
					IDX = nTX + nTm1XX + di*((T-1)*3+1)*XX + dt*3*XX + dk*XX + dx*X + dy;
				}
				assert(i==IDX);
			}

			unsigned int encode(const int x,const int y,const int i,const int j,const int t,const int k) const {

				const int XX = X*X;
				const int nTX = _n*T*X;
				const int nTm1XX = _n*(T-1)*XX;

				if(i!=j && k!=-1){
					return nTX + _n*(T-1)*XX + i*((T-1)*3+1)*XX + t*3*XX + k*XX + x*X + y;
				}

				else if(j==-1 && k==-1){
					return i*T*X + t*X + x;
				}

				else if(i==j && k==-1){
					return nTX + i*(T-1)*XX + t*XX + x*X + y;
				}

				assert(false);

				return 0;
			}
	};
};
#endif
